@@ -47,6 +47,11 @@ typedef void irqreturn_t;
 # define DEVFS_MK_SYMLINK(a,b)
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
+# define io_remap_page_range(vma, vaddr, paddr, size, prot) \
+	remap_pfn_range(vma, vaddr, (paddr) >> PAGE_SHIFT, size, prot)
+#endif
+
 # define my_io_remap_page_range(vma, start, ofs, len, prot) \
 		io_remap_page_range(vma,start,ofs,len,prot)
 
