@@ -19,10 +19,19 @@
 #include <linux/pci.h>
 #include <linux/devfs_fs_kernel.h>
 #include <linux/mm.h>
+#include <linux/thread_info.h>
+#include <linux/smp.h>
+#include <linux/smp_lock.h>
 
 #include <linux/sched.h>
 #include <linux/wait.h>
+#include <linux/syscalls.h>
 
+#define __KERNEL_SYSCALLS__
+#include <linux/unistd.h>
+asmlinkage long (*s_ioperm)(unsigned long from, unsigned long num, int turn_on);
+
+#include <asm/processor.h>
 #include <asm/uaccess.h>
 #include <asm/system.h>   /* cli(), *_flags */
 #include <asm/segment.h>  /* memcpy and such */
@@ -34,7 +43,7 @@
 
 #include "svgalib_helper.h"
 
-#include "../../src/vgaversion.h"
+#include "vgaversion.h"
 
 #include "i810.h"
 #include "interrupt.h"
@@ -42,7 +51,6 @@
 
 int debug=0;
 static int all_devices=0;
-
 int num_devices=0;
 
 static char *sdev_id="svgalib_helper";
@@ -358,6 +366,42 @@ static int svgalib_helper_ioctl( struct inode *inode, struct file *filp,
             }
             
             break;
+#if 0
+		case _IOC_NR(SVGAHELPER_IOPERM):
+			if(minor) return -EPERM;
+#if 1
+			i=0;
+			if(!capable(CAP_SYS_RAWIO)) {
+				i=1;
+				cap_raise(current->cap_effective, CAP_SYS_RAWIO);
+			}
+			s_ioperm=sys_call_table[__NR_ioperm];
+			s_ioperm(0,0x400,1);
+			if(i) {
+				cap_lower(current->cap_effective, CAP_SYS_RAWIO);
+			}
+#else
+			if(!ioperm) {
+				struct thread_struct * t = &current->thread;
+				struct tss_struct * tss;
+				unsigned long *bitmap;
+						
+				ioperm=1;
+				bitmap = kmalloc(IO_BITMAP_BYTES, GFP_KERNEL);
+				if (!bitmap)
+					return -ENOMEM;
+				
+				memset(bitmap, 0, IO_BITMAP_BYTES);
+				t->io_bitmap_ptr = bitmap;
+				tss = &per_cpu(init_tss, get_cpu());
+				t->io_bitmap_max = IO_BITMAP_BYTES;
+				tss->io_bitmap_base = INVALID_IO_BITMAP_OFFSET_LAZY;
+				put_cpu();
+			}
+#endif
+			
+			break;
+#endif
 #ifndef NO_TASK
 		case _IOC_NR(SVGAHELPER_SETDISPLAYSTART):
 			if(minor) {
@@ -714,7 +758,6 @@ void cleanup_module(void)
 	for (i = 0; i < 8; i++)
 		devfs_remove("svga_helper/%d", i);
 	devfs_remove("svga_helper");
-	devfs_remove("svga");
 # endif	
 #endif
 	
