@@ -17,8 +17,68 @@ Rage 128 chipset driver
 #include "vgapci.h"
 #include "endianess.h"
 #include "r128_reg.h"
+#include "xf86PciInfo.h"
+
+static int RADEONProbePLLParameters(void);
+
+static int id;
 
 static enum { Rage128=0, Radeon } chiptype; /* r128io needs to know */
+static enum {
+	CHIP_FAMILY_UNKNOW,
+	CHIP_FAMILY_LEGACY,
+	CHIP_FAMILY_RADEON,
+	CHIP_FAMILY_RV100,
+	CHIP_FAMILY_RS100,    /* U1 (IGP320M) or A3 (IGP320)*/
+	CHIP_FAMILY_RV200,
+	CHIP_FAMILY_RS200,    /* U2 (IGP330M/340M/350M) or A4 (IGP330/340/345/350), RS250 (IGP 7000) */
+	CHIP_FAMILY_R200,
+	CHIP_FAMILY_RV250,
+	CHIP_FAMILY_RS300,    /* RS300/RS350 */
+	CHIP_FAMILY_RV280,
+	CHIP_FAMILY_R300,
+	CHIP_FAMILY_R350,
+	CHIP_FAMILY_RV350,
+	CHIP_FAMILY_RV380,    /* RV370/RV380/M22/M24 */
+	CHIP_FAMILY_R420,     /* R420/R423/M18 */
+	CHIP_FAMILY_LAST
+} ChipFamily; 
+static char *familynames[] = {
+	"Rage 128",
+	"Legacy",
+	"Radeon",
+	"Radeon RV100",
+	"Radeon RS100",
+	"Radeon RV200",
+	"Radeon RS200",
+	"Radeon R200",
+	"Radeon RV250",
+	"Radeon RS300",
+	"Radeon RV280",
+	"Radeon R300",
+	"Radeon R350",
+	"Radeon RV350",
+	"Radeon RV380",
+	"Radeon R420",
+};
+
+#define IS_RV100_VARIANT ((ChipFamily == CHIP_FAMILY_RV100)  ||  \
+		(ChipFamily == CHIP_FAMILY_RV200)  ||  \
+		(ChipFamily == CHIP_FAMILY_RS100)  ||  \
+		(ChipFamily == CHIP_FAMILY_RS200)  ||  \
+		(ChipFamily == CHIP_FAMILY_RV250)  ||  \
+		(ChipFamily == CHIP_FAMILY_RV280)  ||  \
+		(ChipFamily == CHIP_FAMILY_RS300))
+
+#define IS_R300_VARIANT ((ChipFamily == CHIP_FAMILY_R300)  ||  \
+		(ChipFamily == CHIP_FAMILY_RV350) ||  \
+		(ChipFamily == CHIP_FAMILY_R350)  ||  \
+		(ChipFamily == CHIP_FAMILY_RV380) ||  \
+		(ChipFamily == CHIP_FAMILY_R420))
+
+#define TRUE 1
+#define FALSE 0
+
 static int dac6bits;
 #include "r128io.h"
 
@@ -31,7 +91,10 @@ static int dac6bits;
 typedef int Bool;
 
 static int r128_ramtype;
-static int BusCntl, CRTOnly, HasPanelRegs;
+static int BusCntl, CRTOnly, HasPanelRegs, IsIGP, IsMobility, HasSingleDAC, HasCRTC2;
+
+static float mclk, sclk;
+static uint32_t ChipErrata;
 
 typedef struct {
     uint16_t        reference_freq;
@@ -107,6 +170,17 @@ typedef struct {
     uint32_t     ppll_ref_div;
     uint32_t     ppll_div_3;
     uint32_t     htotal_cntl;
+				/* Computed values for PLL2 */
+	uint32_t            dot_clock_freq_2;
+	uint32_t            pll_output_freq_2;
+	int               feedback_div_2;
+	int               post_div_2;
+
+				/* PLL2 registers */
+	uint32_t            p2pll_ref_div;
+	uint32_t            p2pll_div_0;
+	uint32_t            htotal_cntl2;
+
 				/* DDA register */
     uint32_t     dda_config;
     uint32_t     dda_on_off;
@@ -131,6 +205,7 @@ typedef struct {        /* All values in XCLKS    */
 } R128RAMRec, *R128RAMPtr;
 
 #define R128_TOTAL_REGS (VGA_TOTAL_REGS + sizeof(R128SaveRec))
+
 static int R128MinBits(int val)
 {
     int bits;
@@ -139,10 +214,12 @@ static int R128MinBits(int val)
     for (bits = 0; val; val >>= 1, ++bits);
     return bits;
 }
+
 static int R128Div(int n, int d)
 {
     return (n + (d / 2)) / d;
 }
+
 static R128PLLRec pll;
 
 static R128RAMRec ram[] = {        /* Memory Specifications
@@ -154,10 +231,49 @@ static R128RAMRec ram[] = {        /* Memory Specifications
     { 4, 4, 1, 2, 1, 2, 1, 16, 12, "64-bit SDR SGRAM 2:1" },
     { 4, 4, 3, 3, 2, 3, 1, 16, 12, "64-bit DDR SGRAM" },
 };
+void RADEONPllErrataAfterIndex(void)
+{
+	if (!(ChipErrata & CHIP_ERRATA_PLL_DUMMYREADS))
+		return;
+
+	/* This workaround is necessary on rv200 and RS200 or PLL
+	 * reads may return garbage (among others...)
+	 */
+	(void)INREG(R128_CLOCK_CNTL_DATA);
+	(void)INREG(R128_CRTC_GEN_CNTL);
+}
+
+void RADEONPllErrataAfterData(void)
+{
+    /* This workarounds is necessary on RV100, RS100 and RS200 chips
+     * or the chip could hang on a subsequent access
+	 */
+
+	if (ChipErrata & CHIP_ERRATA_PLL_DELAY) {
+		/* we can't deal with posted writes here ... */
+		usleep(5000);
+	}
+
+			    /* This function is required to workaround a hardware bug in
+				 * some (all?) revisions of the R300.  This workaround should 
+				 * be called after every CLOCK_CNTL_INDEX register access.  
+				 * If not, register reads afterward may not be correct.
+				 */
+
+	if (ChipErrata & CHIP_ERRATA_R300_CG) {
+		uint32_t save, tmp;
+		
+		save = INREG(R128_CLOCK_CNTL_INDEX);
+		tmp = save & ~(0x3f | R128_PLL_WR_EN);
+		OUTREG(R128_CLOCK_CNTL_INDEX, tmp);
+		tmp = INREG(R128_CLOCK_CNTL_DATA);
+		OUTREG(R128_CLOCK_CNTL_INDEX, save);
+	}
+}
 
 static unsigned R128INPLL(int addr)
 {
-    OUTREG8(R128_CLOCK_CNTL_INDEX, addr & 0x1f);
+    OUTREG8(R128_CLOCK_CNTL_INDEX, addr & 0x3f);
     return INREG(R128_CLOCK_CNTL_DATA);
 }
 
@@ -284,7 +400,8 @@ static void R128PLLWaitForReadUpdateComplete(void)
 
 static void R128PLLWriteUpdate(void)
 {
-    OUTPLLP(R128_PPLL_REF_DIV, R128_PPLL_ATOMIC_UPDATE_W, 0xffff);
+	while (INPLL ( R128_PPLL_REF_DIV) & R128_PPLL_ATOMIC_UPDATE_R);
+    OUTPLLP(R128_PPLL_REF_DIV, R128_PPLL_ATOMIC_UPDATE_W, ~(R128_PPLL_ATOMIC_UPDATE_W));
 }
 
 /* Write PLL registers. */
@@ -306,10 +423,30 @@ static void RADEONRestorePLLRegisters(R128SavePtr restore)
 	    R128_PLL_DIV_SEL,
 	    ~(R128_PLL_DIV_SEL));
 
-    OUTPLLP(R128_PPLL_REF_DIV,
-	    restore->ppll_ref_div,
-	    ~R128_PPLL_REF_DIV_MASK);
 
+	if(ChipFamily>=CHIP_FAMILY_R300) /* IS_R300_VARIANT */
+	{
+		if (restore->ppll_ref_div & R300_PPLL_REF_DIV_ACC_MASK) {
+			/* When restoring console mode, use saved PPLL_REF_DIV
+			 * setting. */
+			OUTPLLP(R128_PPLL_REF_DIV,
+					restore->ppll_ref_div,
+					0);
+		} else {
+		   	/* R300 uses ref_div_acc field as real ref divider */
+	 		OUTPLLP(R128_PPLL_REF_DIV,
+					(restore->ppll_ref_div << R300_PPLL_REF_DIV_ACC_SHIFT),
+   					~R300_PPLL_REF_DIV_ACC_MASK);
+		}
+	} else {
+
+		OUTPLLP(R128_PPLL_REF_DIV,
+				restore->ppll_ref_div,
+				~R128_PPLL_REF_DIV_MASK);
+	}
+
+	
+	
     OUTPLLP(R128_PPLL_DIV_3,
 	    restore->ppll_div_3,
 	    ~R128_PPLL_FB3_DIV_MASK);
@@ -330,7 +467,7 @@ static void RADEONRestorePLLRegisters(R128SavePtr restore)
 	      | R128_PPLL_ATOMIC_UPDATE_EN
 	      | R128_PPLL_VGA_ATOMIC_UPDATE_EN));
 
-    usleep(5000); /* Let the clock to lock */
+    usleep(500000); /* Let the clock to lock */
 
     OUTPLLP(R128_VCLK_ECP_CNTL,
 	    RADEON_VCLK_SRC_SEL_PPLLCLK,
@@ -419,8 +556,9 @@ static void R128RestoreMode(R128SavePtr restore)
     }
             
     if(chiptype == Rage128) {
-	R128RestoreDDARegisters(restore);
+		R128RestoreDDARegisters(restore);
     }
+
     R128RestorePalette(restore);
 }
 
@@ -602,13 +740,14 @@ static Bool R128InitCrtcRegisters(R128SavePtr save,
 	return 0;
     }
     
-    if (HasPanelRegs)
-	if (CRTOnly) hsync_fudge = hsync_fudge_fp_crt[format-1];
-	else               hsync_fudge = hsync_fudge_fp[format-1];
-    else
-                           hsync_fudge = hsync_fudge_default[format-1];
-
-    save->crtc_gen_cntl = (R128_CRTC_EXT_DISP_EN
+	if(chiptype==Rage128) {
+		if (HasPanelRegs)
+			if (CRTOnly) hsync_fudge = hsync_fudge_fp_crt[format-1];
+			else hsync_fudge = hsync_fudge_fp[format-1];
+		else hsync_fudge = hsync_fudge_default[format-1];
+	} else hsync_fudge = 0;
+	
+	save->crtc_gen_cntl = (R128_CRTC_EXT_DISP_EN
 			  | R128_CRTC_EN
 			  | (format << 8)
 			  | ((mode->flags & DOUBLESCAN)
@@ -624,8 +763,8 @@ static Bool R128InitCrtcRegisters(R128SavePtr save,
 			   | R128_DAC_VGA_ADR_EN
 			   | (dac6bits ? 0 : R128_DAC_8BIT_EN));
 
-    save->crtc_h_total_disp = ((((mode->CrtcHTotal / 8) - 1) & 0xffff)
-			      | (((mode->CrtcHDisplay / 8) - 1) << 16));
+    save->crtc_h_total_disp = ((((mode->CrtcHTotal / 8) - 1) & 0x3ff)
+			      | ((((mode->CrtcHDisplay / 8) - 1) & 0x1ff) << 16));
 
     hsync_wid = (mode->CrtcHSyncEnd - mode->CrtcHSyncStart) / 8;
     if (!hsync_wid)       hsync_wid = 1;
@@ -633,8 +772,8 @@ static Bool R128InitCrtcRegisters(R128SavePtr save,
 
     hsync_start = mode->CrtcHSyncStart - 8 + hsync_fudge;
 
-    save->crtc_h_sync_strt_wid = ((hsync_start & 0xfff)
-				 | (hsync_wid << 16)
+    save->crtc_h_sync_strt_wid = ((hsync_start & 0x1fff)
+				 | ((hsync_wid & 0x3f) << 16)
 				 | ((mode->flags & NHSYNC)
 				    ? R128_CRTC_H_SYNC_POL
 				    : 0));
@@ -654,7 +793,11 @@ static Bool R128InitCrtcRegisters(R128SavePtr save,
 
     vsync_wid = mode->CrtcVSyncEnd - mode->CrtcVSyncStart;
     if (!vsync_wid)       vsync_wid = 1;
-    if (vsync_wid > 0x1f) vsync_wid = 0x1f;
+	if(chiptype==Rage128) {
+		if (vsync_wid > 0x1f) vsync_wid = 0x1f;
+	} else {
+		vsync_wid &= 0x1f;
+	}
 
     save->crtc_v_sync_strt_wid = (((mode->CrtcVSyncStart - 1) & 0xfff)
 				 | (vsync_wid << 16)
@@ -662,12 +805,22 @@ static Bool R128InitCrtcRegisters(R128SavePtr save,
 				    ? R128_CRTC_V_SYNC_POL
 				    : 0));
     save->crtc_offset      = 0;
+
+#if 0
 	/* The register reference says this bit should be 0,
 	 * but on my RV280 it appears as if it should be 1
 	 * for changing offset at vertical blank. */
     save->crtc_offset_cntl = 1<<16; 
+#endif
+	save->crtc_offset_cntl = INREG(R128_CRTC_OFFSET_CNTL);
 
-	save->crtc_pitch       = info->width / 8;
+	if(chiptype==Rage128) {
+		save->crtc_pitch       = info->width / 8;
+	} else {
+		save->crtc_pitch  = (((info->width * info->bitsPerPixel) +
+				((info->bitsPerPixel * 8) -1)) /
+				(info->bitsPerPixel * 8));
+	}
 
     save->config_cntl |= R128_CFG_VGA_RAM_EN;
     
@@ -797,37 +950,43 @@ static void R128InitFPRegisters(R128SavePtr orig, R128SavePtr save,
 static void RADEONInitPLLRegisters(R128SavePtr save, R128PLLPtr pll,
 				 double dot_clock)
 {
-    unsigned int freq = dot_clock * 100;
-    struct {
-	int divider;
-	int bitvalue;
-    } *post_div,
-      post_divs[]   = {
+    unsigned int freq; 
+	
+	struct {
+		int divider;
+		int bitvalue;
+    } *post_div, post_divs[]   = {
 				/* From RAGE 128 VR/RAGE 128 GL Register
 				   Reference Manual (Technical Reference
 				   Manual P/N RRG-G04100-C Rev. 0.04), page
 				   3-17 (PLL_DIV_[3:0]).  */
-	{  1, 0 },              /* VCLK_SRC                 */
-	{  2, 1 },              /* VCLK_SRC/2               */
-	{  4, 2 },              /* VCLK_SRC/4               */
-	{  8, 3 },              /* VCLK_SRC/8               */
-
-	{  3, 4 },              /* VCLK_SRC/3               */
-	{ 16, 5 },
-        {  6, 6 },              /* VCLK_SRC/6               */
-	{ 12, 7 },              /* VCLK_SRC/12              */
-	{  0, 0 }
+		{  1, 0 },              /* VCLK_SRC                 */
+		{  2, 1 },              /* VCLK_SRC/2               */
+		{  4, 2 },              /* VCLK_SRC/4               */
+		{  8, 3 },              /* VCLK_SRC/8               */
+		{  3, 4 },              /* VCLK_SRC/3               */
+		{ 16, 5 },
+		{  6, 6 },              /* VCLK_SRC/6               */
+		{ 12, 7 },              /* VCLK_SRC/12              */
+		{  0, 0 }
     };
+	
+	freq = dot_clock * 100;
 
     if (freq > pll->max_pll_freq)      freq = pll->max_pll_freq;
     if (freq * 12 < pll->min_pll_freq) freq = pll->min_pll_freq / 12;
 
     for (post_div = &post_divs[0]; post_div->divider; ++post_div) {
-	save->pll_output_freq = post_div->divider * freq;
-	if (save->pll_output_freq >= pll->min_pll_freq
-	    && save->pll_output_freq <= pll->max_pll_freq) break;
+		save->pll_output_freq = post_div->divider * freq;
+		if (save->pll_output_freq >= pll->min_pll_freq
+			&& save->pll_output_freq <= pll->max_pll_freq) break;
     }
 
+	if (!post_div->divider) {
+  		save->pll_output_freq = freq;
+  		post_div = &post_divs[0];
+ 	}
+
     save->dot_clock_freq = freq;
     save->feedback_div   = R128Div(pll->reference_div * save->pll_output_freq,
 				   pll->reference_freq);
@@ -843,33 +1002,31 @@ static void R128InitPLLRegisters(R128SavePtr save, R128PLLPtr pll,
 {
     unsigned int freq = dot_clock * 100;
     struct {
-	int divider;
-	int bitvalue;
-    } *post_div,
-      post_divs[]   = {
+		int divider;
+		int bitvalue;
+    } *post_div, post_divs[]   = {
 				/* From RAGE 128 VR/RAGE 128 GL Register
 				   Reference Manual (Technical Reference
 				   Manual P/N RRG-G04100-C Rev. 0.04), page
 				   3-17 (PLL_DIV_[3:0]).  */
-	{  1, 0 },              /* VCLK_SRC                 */
-	{  2, 1 },              /* VCLK_SRC/2               */
-	{  4, 2 },              /* VCLK_SRC/4               */
-	{  8, 3 },              /* VCLK_SRC/8               */
-
-	{  3, 4 },              /* VCLK_SRC/3               */
-				/* bitvalue = 5 is reserved */
-	{  6, 6 },              /* VCLK_SRC/6               */
-	{ 12, 7 },              /* VCLK_SRC/12              */
-	{  0, 0 }
+		{  1, 0 },              /* VCLK_SRC                 */
+		{  2, 1 },              /* VCLK_SRC/2               */
+		{  4, 2 },              /* VCLK_SRC/4               */
+		{  8, 3 },              /* VCLK_SRC/8               */
+		{  3, 4 },              /* VCLK_SRC/3               */
+					/* bitvalue = 5 is reserved */
+		{  6, 6 },              /* VCLK_SRC/6               */
+		{ 12, 7 },              /* VCLK_SRC/12              */
+		{  0, 0 }
     };
 
     if (freq > pll->max_pll_freq)      freq = pll->max_pll_freq;
     if (freq * 12 < pll->min_pll_freq) freq = pll->min_pll_freq / 12;
 
     for (post_div = &post_divs[0]; post_div->divider; ++post_div) {
-	save->pll_output_freq = post_div->divider * freq;
-	if (save->pll_output_freq >= pll->min_pll_freq
-	    && save->pll_output_freq <= pll->max_pll_freq) break;
+		save->pll_output_freq = post_div->divider * freq;
+		if (save->pll_output_freq >= pll->min_pll_freq
+			&& save->pll_output_freq <= pll->max_pll_freq) break;
     }
 
     save->dot_clock_freq = freq;
@@ -946,7 +1103,9 @@ static void R128InitPalette(R128SavePtr save)
 /* Define registers for a requested video mode. */
 static Bool R128Init(ModeTiming *mode, ModeInfo *info, R128SavePtr save)
 {
-    double      dot_clock = mode->pixelClock/1000.0;
+    double dot_clock;
+
+	dot_clock = mode->pixelClock/1000.0;
 
     R128InitCommonRegisters(save);
     if (!R128InitCrtcRegisters(save, mode, info)) return 0;
@@ -963,8 +1122,10 @@ static Bool R128Init(ModeTiming *mode, ModeInfo *info, R128SavePtr save)
             break;
     }
     
-    if (!R128InitDDARegisters(save, &pll, info))
-	return 0;
+	if(chiptype == Rage128) {
+		if (!R128InitDDARegisters(save, &pll, info))
+			return 0;
+	}
 //    if (!info->PaletteSavedOnVT) 
     R128InitPalette(save);
 
@@ -1302,9 +1463,8 @@ DriverSpecs __svgalib_r128_driverspecs =
 static int r128_init(int force, int par1, int par2)
 {
     unsigned int buf[64];
-    int found=0, id, NOBIOS=0;
+    int found=0, NOBIOS=0;
     unsigned char *BIOS_POINTER;
-    char chipnames[2][9]={"Rage 128", "Radeon"};
 
     r128_memory=0;
 	chiptype=-1;
@@ -1351,7 +1511,8 @@ static int r128_init(int force, int par1, int par2)
         ((id>>8)==0x59) ||
         ((id>>8)==0x5b) ||
         ((id>>8)==0x5c) ||
-        ((id>>8)==0x5d)
+        ((id>>8)==0x5d) ||
+        ((id>>8)==0x5e)
 		) chiptype = Radeon;
 
 	if( (id == 0x4158) ||
@@ -1360,6 +1521,231 @@ static int r128_init(int force, int par1, int par2)
 	
     if(chiptype==-1) return 1;
     
+	ChipFamily = CHIP_FAMILY_LEGACY;
+	if(chiptype == Radeon) {
+		switch(id) {
+			case PCI_CHIP_RADEON_LY:
+			case PCI_CHIP_RADEON_LZ:
+				IsMobility = TRUE;
+				ChipFamily = CHIP_FAMILY_RV100;
+				break;
+				
+			case PCI_CHIP_RV100_QY:
+			case PCI_CHIP_RV100_QZ:
+			case PCI_CHIP_RN50_515E:  /* RN50 is based on the RV100 but 3D isn't guaranteed to work.  YMMV. */
+			case PCI_CHIP_RN50_5969:
+				ChipFamily = CHIP_FAMILY_RV100;
+				
+				break;
+				
+			case PCI_CHIP_RS100_4336:
+				IsMobility = TRUE;
+			case PCI_CHIP_RS100_4136:
+				ChipFamily = CHIP_FAMILY_RS100;
+				IsIGP = TRUE;
+				break;
+				
+			case PCI_CHIP_RS200_4337:
+				IsMobility = TRUE;
+			case PCI_CHIP_RS200_4137:
+				ChipFamily = CHIP_FAMILY_RS200;
+				IsIGP = TRUE;
+				break;
+				
+			case PCI_CHIP_RS250_4437:
+				IsMobility = TRUE;
+			case PCI_CHIP_RS250_4237:
+				ChipFamily = CHIP_FAMILY_RS200;
+				IsIGP = TRUE;
+				break;
+				
+			case PCI_CHIP_R200_BB:
+			case PCI_CHIP_R200_BC:
+			case PCI_CHIP_R200_QH:
+			case PCI_CHIP_R200_QL:
+			case PCI_CHIP_R200_QM:
+				ChipFamily = CHIP_FAMILY_R200;
+				break;
+				
+			case PCI_CHIP_RADEON_LW:
+			case PCI_CHIP_RADEON_LX:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV200_QW: /* RV200 desktop */
+			case PCI_CHIP_RV200_QX:
+				ChipFamily = CHIP_FAMILY_RV200;
+				break;
+				
+			case PCI_CHIP_RV250_Ld:
+			case PCI_CHIP_RV250_Lf:
+			case PCI_CHIP_RV250_Lg:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV250_If:
+			case PCI_CHIP_RV250_Ig:
+				ChipFamily = CHIP_FAMILY_RV250;
+				break;
+				
+			case PCI_CHIP_RS300_5835:
+			case PCI_CHIP_RS350_7835:
+				IsMobility = TRUE;
+			case PCI_CHIP_RS300_5834:
+			case PCI_CHIP_RS350_7834:
+				ChipFamily = CHIP_FAMILY_RS300;
+				IsIGP = TRUE;
+				HasSingleDAC = TRUE;
+				break;
+				
+			case PCI_CHIP_RV280_5C61:
+			case PCI_CHIP_RV280_5C63:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV280_5960:
+			case PCI_CHIP_RV280_5961:
+			case PCI_CHIP_RV280_5962:
+			case PCI_CHIP_RV280_5964:
+			case PCI_CHIP_RV280_5965:
+				ChipFamily = CHIP_FAMILY_RV280;
+				break;
+				
+			case PCI_CHIP_R300_AD:
+			case PCI_CHIP_R300_AE:
+			case PCI_CHIP_R300_AF:
+			case PCI_CHIP_R300_AG:
+			case PCI_CHIP_R300_ND:
+			case PCI_CHIP_R300_NE:
+			case PCI_CHIP_R300_NF:
+			case PCI_CHIP_R300_NG:
+				ChipFamily = CHIP_FAMILY_R300;
+				break;
+				
+			case PCI_CHIP_RV350_NP:
+			case PCI_CHIP_RV350_NQ:
+			case PCI_CHIP_RV350_NR:
+			case PCI_CHIP_RV350_NS:
+			case PCI_CHIP_RV350_NT:
+			case PCI_CHIP_RV350_NV:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV350_AP:
+			case PCI_CHIP_RV350_AQ:
+			case PCI_CHIP_RV360_AR:
+			case PCI_CHIP_RV350_AS:
+			case PCI_CHIP_RV350_AT:
+			case PCI_CHIP_RV350_AV:
+			case PCI_CHIP_RV350_4155:
+				ChipFamily = CHIP_FAMILY_RV350;
+				break;
+				
+			case PCI_CHIP_R350_AH:
+			case PCI_CHIP_R350_AI:
+			case PCI_CHIP_R350_AJ:
+			case PCI_CHIP_R350_AK:
+			case PCI_CHIP_R350_NH:
+			case PCI_CHIP_R350_NI:
+			case PCI_CHIP_R350_NK:
+			case PCI_CHIP_R360_NJ:
+				ChipFamily = CHIP_FAMILY_R350;
+				break;
+				
+			case PCI_CHIP_RV380_3150:
+			case PCI_CHIP_RV380_3154:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV380_3E50:
+			case PCI_CHIP_RV380_3E54:
+				ChipFamily = CHIP_FAMILY_RV380;
+				break;
+				
+			case PCI_CHIP_RV370_5460:
+			case PCI_CHIP_RV370_5464:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV370_5B60:
+			case PCI_CHIP_RV370_5B62:
+			case PCI_CHIP_RV370_5B64:
+			case PCI_CHIP_RV370_5B65:
+				ChipFamily = CHIP_FAMILY_RV380;
+				break;
+				
+			case PCI_CHIP_RS400_5A42:
+			case PCI_CHIP_RC410_5A62:
+			case PCI_CHIP_RS480_5955:
+			case PCI_CHIP_RS482_5975:
+				IsMobility = TRUE;
+			case PCI_CHIP_RS400_5A41:
+			case PCI_CHIP_RC410_5A61:
+			case PCI_CHIP_RS480_5954:
+			case PCI_CHIP_RS482_5974:
+				ChipFamily = CHIP_FAMILY_RV380; /*CHIP_FAMILY_RS400*/
+				/*IsIGP = TRUE;*/ /* ??? */
+				/*HasSingleDAC = TRUE;*/ /* ??? */
+				break;
+				
+			case PCI_CHIP_RV410_564A:
+			case PCI_CHIP_RV410_564B:
+			case PCI_CHIP_RV410_5652:
+			case PCI_CHIP_RV410_5653:
+				IsMobility = TRUE;
+			case PCI_CHIP_RV410_5E48:
+			case PCI_CHIP_RV410_5E4B:
+			case PCI_CHIP_RV410_5E4A:
+			case PCI_CHIP_RV410_5E4D:
+			case PCI_CHIP_RV410_5E4C:
+			case PCI_CHIP_RV410_5E4F:
+				ChipFamily = CHIP_FAMILY_R420; /* CHIP_FAMILY_RV410*/
+				break;
+				
+			case PCI_CHIP_R420_JN:
+				IsMobility = TRUE;
+			case PCI_CHIP_R420_JH:
+			case PCI_CHIP_R420_JI:
+			case PCI_CHIP_R420_JJ:
+			case PCI_CHIP_R420_JK:
+			case PCI_CHIP_R420_JL:
+			case PCI_CHIP_R420_JM:
+			case PCI_CHIP_R420_JP:
+			case PCI_CHIP_R420_4A4F:
+				ChipFamily = CHIP_FAMILY_R420;
+				break;
+				
+			case PCI_CHIP_R423_UH:
+			case PCI_CHIP_R423_UI:
+			case PCI_CHIP_R423_UJ:
+			case PCI_CHIP_R423_UK:
+			case PCI_CHIP_R423_UQ:
+			case PCI_CHIP_R423_UR:
+			case PCI_CHIP_R423_UT:
+			case PCI_CHIP_R423_5D57:
+			case PCI_CHIP_R423_5550:
+				ChipFamily = CHIP_FAMILY_R420;
+				break;
+				
+			case PCI_CHIP_R430_5D49:
+			case PCI_CHIP_R430_5D4A:
+			case PCI_CHIP_R430_5D48:
+				IsMobility = TRUE;
+			case PCI_CHIP_R430_554F:
+			case PCI_CHIP_R430_554D:
+			case PCI_CHIP_R430_554E:
+			case PCI_CHIP_R430_554C:
+				ChipFamily = CHIP_FAMILY_R420; /*CHIP_FAMILY_R430*/
+				break;
+				
+			case PCI_CHIP_R480_5D4C:
+			case PCI_CHIP_R480_5D50:
+			case PCI_CHIP_R480_5D4E:
+			case PCI_CHIP_R480_5D4F:
+			case PCI_CHIP_R480_5D52:
+			case PCI_CHIP_R480_5D4D:
+			case PCI_CHIP_R481_4B4B:
+			case PCI_CHIP_R481_4B4A:
+			case PCI_CHIP_R481_4B49:
+			case PCI_CHIP_R481_4B4C:
+				ChipFamily = CHIP_FAMILY_R420; /*CHIP_FAMILY_R480*/
+				break;
+				
+			default:
+				/* Original Radeon/7200 */
+				ChipFamily = CHIP_FAMILY_RADEON;
+				HasCRTC2 = FALSE;
+		}
+	}
+
     r128_linear_base=buf[4]&0xffffff00;
     r128_mmio_base=buf[6]&0xffffff00;
 	
@@ -1374,13 +1760,49 @@ static int r128_init(int force, int par1, int par2)
     CRTOnly		= 1;
     r128_ramtype	= 1;
 
+	ChipErrata = 0;
+
+	if (ChipFamily == CHIP_FAMILY_R300 &&
+			(INREG(RADEON_CONFIG_CNTL) & RADEON_CFG_ATI_REV_ID_MASK)
+			== RADEON_CFG_ATI_REV_A11)
+		ChipErrata |= CHIP_ERRATA_R300_CG;
+	
+	if (ChipFamily == CHIP_FAMILY_RV200 ||
+		ChipFamily == CHIP_FAMILY_RS200)
+		ChipErrata |= CHIP_ERRATA_PLL_DUMMYREADS;
+	
+	if (ChipFamily == CHIP_FAMILY_RV100 ||
+		ChipFamily == CHIP_FAMILY_RS100 ||
+		ChipFamily == CHIP_FAMILY_RS200)
+		ChipErrata |= CHIP_ERRATA_PLL_DELAY;
+
+//	RADEONProbePLLParameters();
+		
 #ifdef __i386__
     if(__svgalib_secondary) 
 #endif
-        NOBIOS=1;
+		NOBIOS=1;
+	
+	pll.min_pll_freq = 0;
 
-    if(NOBIOS) {
-        int x_mpll_ref_fb_div, xclk_cntl, Nx, M;
+#define R128_BIOS16(x) (*(unsigned short *)(BIOS_POINTER + x))
+#define R128_BIOS32(x) (*(unsigned int *)(BIOS_POINTER + x))
+    if(!NOBIOS) {
+        uint16_t bios_header;
+        uint16_t pll_info_block;
+        BIOS_POINTER = mmap(0, 64*1024, PROT_READ | PROT_WRITE, MAP_SHARED, __svgalib_mem_fd,
+       			    0xc0000);
+        bios_header    = R128_BIOS16(0x48);
+        pll_info_block = R128_BIOS16(bios_header + 0x30);
+        pll.reference_freq = R128_BIOS16(pll_info_block + 0x0e);
+        pll.reference_div  = R128_BIOS16(pll_info_block + 0x10);
+        pll.min_pll_freq   = R128_BIOS32(pll_info_block + 0x12);
+        pll.max_pll_freq   = R128_BIOS32(pll_info_block + 0x16);
+        pll.xclk           = R128_BIOS16(pll_info_block + 0x08);
+        munmap(BIOS_POINTER, 64*1024);
+    }
+    if((pll.min_pll_freq<2500) || (pll.min_pll_freq>1000000)) {
+        int x_mpll_ref_fb_div, xclk_cntl, Nx, M, tmp;
         int PostDivSet[] = {0, 1, 2, 4, 8, 3, 6, 12};
     
         switch(chiptype) {
@@ -1398,30 +1820,26 @@ static int r128_init(int force, int par1, int par2)
                                       (M * PostDivSet[xclk_cntl]));
                 break;
             case Radeon:
-                pll.reference_freq = 2700;
-                pll.min_pll_freq   = 12500;
+                pll.reference_freq = IsIGP ? 1432 : 2700;
+                pll.min_pll_freq   = 20000; /* X says 12500 */
                 pll.max_pll_freq   = 35000;
                 pll.reference_div  = 67;
-                pll.xclk 	   = 16667;
+				if(ChipFamily==CHIP_FAMILY_R420) {
+					pll.min_pll_freq   = 20000;
+					pll.max_pll_freq   = 50000;
+				}
+				tmp = INPLL(R128_PPLL_REF_DIV);
+				if (IS_R300_VARIANT || (ChipFamily == CHIP_FAMILY_RS300)) {
+					pll.reference_div = (tmp & R300_PPLL_REF_DIV_ACC_MASK) >> 
+						R300_PPLL_REF_DIV_ACC_SHIFT;
+				} else {
+					pll.reference_div = tmp & R128_PPLL_REF_DIV_MASK;
+				}				
+				if (pll.reference_div < 2) pll.reference_div = 12;
+										
                 break;
         }
-    } else 
-#define R128_BIOS16(x) (*(unsigned short *)(BIOS_POINTER + x))
-#define R128_BIOS32(x) (*(unsigned int *)(BIOS_POINTER + x))
-    {
-        uint16_t bios_header;
-        uint16_t pll_info_block;
-        BIOS_POINTER = mmap(0, 64*1024, PROT_READ | PROT_WRITE, MAP_SHARED, __svgalib_mem_fd,
-       			    0xc0000);
-        bios_header    = R128_BIOS16(0x48);
-        pll_info_block = R128_BIOS16(bios_header + 0x30);
-        pll.reference_freq = R128_BIOS16(pll_info_block + 0x0e);
-        pll.reference_div  = R128_BIOS16(pll_info_block + 0x10);
-        pll.min_pll_freq   = R128_BIOS32(pll_info_block + 0x12);
-        pll.max_pll_freq   = R128_BIOS32(pll_info_block + 0x16);
-        pll.xclk           = R128_BIOS16(pll_info_block + 0x08);
-        munmap(BIOS_POINTER, 64*1024);
-    }
+    } 
 #if 0
 fprintf(stderr,"pll: %i %i %i %i %i\n",pll.reference_freq,pll.reference_div,
     pll.min_pll_freq,    pll.max_pll_freq, pll.xclk);
@@ -1430,7 +1848,8 @@ fprintf(stderr,"pll: %i %i %i %i %i\n",pll.reference_freq,pll.reference_div,
     r128_mapio();
 	
     if (__svgalib_driver_report) {
-		fprintf(stderr,"Using ATI %s driver, %iKB.\n",chipnames[chiptype],r128_memory);
+		fprintf(stderr,"Using ATI R128/Radeon driver, %s with %iMB found.\n",
+				familynames[chiptype*ChipFamily],r128_memory/1024);
     };
 
 	__svgalib_modeinfo_linearset = LINEAR_CAN;
@@ -1461,3 +1880,222 @@ fprintf(stderr,"pll: %i %i %i %i %i\n",pll.reference_freq,pll.reference_div,
 
     return 0;
 }
+
+static void xf86getsecs(long * secs, long * usecs)
+{
+	struct timeval tv;
+	
+	gettimeofday(&tv, NULL);
+	if (secs)
+		*secs = tv.tv_sec;
+	if (usecs)
+		*usecs= tv.tv_usec;
+	
+	return;
+}
+
+#define xf86DrvMsg(a,b, ...) printf(__VA_ARGS__)
+
+static int RADEONProbePLLParameters(void)
+{
+    R128PLLRec  ppll; 
+	R128PLLPtr   pll = &ppll;
+    unsigned char ppll_div_sel;
+    unsigned mpll_fb_div, spll_fb_div, M;
+    unsigned xclk, tmp, ref_div;
+    int hTotal, vTotal, num, denom, m, n;
+    float hz, prev_xtal, vclk, xtal, mpll, spll;
+    long start_secs, start_usecs, stop_secs, stop_usecs, total_usecs;
+    long to1_secs, to1_usecs, to2_secs, to2_usecs;
+    unsigned int f1, f2, f3;
+    int tries = 0;
+
+    prev_xtal = 0;
+ again:
+    xtal = 0;
+    if (++tries > 10)
+           goto failed;
+
+    xf86getsecs(&to1_secs, &to1_usecs);
+    f1 = INREG(RADEON_CRTC_CRNT_FRAME);
+    for (;;) {
+       f2 = INREG(RADEON_CRTC_CRNT_FRAME);
+       if (f1 != f2)
+	    break;
+       xf86getsecs(&to2_secs, &to2_usecs);
+       if ((to2_secs - to1_secs) > 1) {
+           xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Clock not counting...\n");
+           goto failed;
+       }
+    }
+    xf86getsecs(&start_secs, &start_usecs);
+    for(;;) {
+       f3 = INREG(RADEON_CRTC_CRNT_FRAME);
+       if (f3 != f2)
+	    break;
+       xf86getsecs(&to2_secs, &to2_usecs);
+       if ((to2_secs - start_secs) > 1)
+           goto failed;
+    }
+    xf86getsecs(&stop_secs, &stop_usecs);
+
+    if ((stop_secs - start_secs) != 0)
+           goto again;
+    total_usecs = abs(stop_usecs - start_usecs);
+    if (total_usecs == 0)
+           goto again;
+    hz = 1000000.0/(float)total_usecs;
+
+    hTotal = ((INREG(R128_CRTC_H_TOTAL_DISP) & 0x3ff) + 1) * 8;
+    vTotal = ((INREG(R128_CRTC_V_TOTAL_DISP) & 0xfff) + 1);
+    vclk = (float)(hTotal * (float)(vTotal * hz));
+
+    switch((INPLL( R128_PPLL_REF_DIV) & 0x30000) >> 16) {
+    case 0:
+    default:
+        num = 1;
+        denom = 1;
+        break;
+    case 1:
+        n = ((INPLL( R128_X_MPLL_REF_FB_DIV) >> 16) & 0xff);
+        m = (INPLL( R128_X_MPLL_REF_FB_DIV) & 0xff);
+        num = 2*n;
+        denom = 2*m;
+        break;
+    case 2:
+        n = ((INPLL( R128_X_MPLL_REF_FB_DIV) >> 8) & 0xff);
+        m = (INPLL( R128_X_MPLL_REF_FB_DIV) & 0xff);
+        num = 2*n;
+        denom = 2*m;
+        break;
+     }
+
+    ppll_div_sel = INREG8(R128_CLOCK_CNTL_INDEX + 1) & 0x3;
+    RADEONPllErrataAfterIndex();
+
+    n = (INPLL( R128_PPLL_DIV_0 + ppll_div_sel) & 0x7ff);
+    m = (INPLL( R128_PPLL_REF_DIV) & 0x3ff);
+
+    num *= n;
+    denom *= m;
+
+    switch ((INPLL( R128_PPLL_DIV_0 + ppll_div_sel) >> 16) & 0x7) {
+    case 1:
+        denom *= 2;
+        break;
+    case 2:
+        denom *= 4;
+        break;
+    case 3:
+        denom *= 8;
+        break;
+    case 4:
+        denom *= 3;
+        break;
+    case 6:
+        denom *= 6;
+        break;
+    case 7:
+        denom *= 12;
+        break;
+    }
+
+    xtal = (int)(vclk *(float)denom/(float)num);
+
+    if ((xtal > 26900000) && (xtal < 27100000))
+        xtal = 2700;
+    else if ((xtal > 14200000) && (xtal < 14400000))
+        xtal = 1432;
+    else if ((xtal > 29400000) && (xtal < 29600000))
+        xtal = 2950;
+    else
+       goto again;
+ failed:
+    if (xtal == 0) {
+       xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Failed to probe xtal value ! "
+                  "Using default 27Mhz\n");
+       xtal = 2700;
+    } else {
+       if (prev_xtal == 0) {
+           prev_xtal = xtal;
+           tries = 0;
+           goto again;
+       } else if (prev_xtal != xtal) {
+           prev_xtal = 0;
+           goto again;
+       }
+    }
+
+    tmp = INPLL( R128_X_MPLL_REF_FB_DIV);
+    ref_div = INPLL( R128_PPLL_REF_DIV) & 0x3ff;
+
+    /* Some sanity check based on the BIOS code .... */
+#if 0
+    if (ref_div < 2) {
+       CARD32 tmp;
+       tmp = INPLL( R128_PPLL_REF_DIV);
+       if (IS_R300_VARIANT || (info->ChipFamily == CHIP_FAMILY_RS300))
+           ref_div = (tmp & R300_PPLL_REF_DIV_ACC_MASK) >>
+                   R300_PPLL_REF_DIV_ACC_SHIFT;
+       else
+           ref_div = tmp & R128_PPLL_REF_DIV_MASK;
+       if (ref_div < 2)
+           ref_div = 12;
+    }
+#endif
+    /* Calculate "base" xclk straight from MPLL, though that isn't
+     * really useful (hopefully). This isn't called XCLK anymore on
+     * radeon's...
+     */
+    mpll_fb_div = (tmp & 0xff00) >> 8;
+    spll_fb_div = (tmp & 0xff0000) >> 16;
+    M = (tmp & 0xff);
+    xclk = R128Div((2 * mpll_fb_div * xtal), (M));
+
+    /*
+     * Calculate MCLK based on MCLK-A
+     */
+    mpll = (2.0 * (float)mpll_fb_div * (xtal / 100.0)) / (float)M;
+    spll = (2.0 * (float)spll_fb_div * (xtal / 100.0)) / (float)M;
+
+    tmp = INPLL( R128_MCLK_CNTL) & 0x7;
+    switch(tmp) {
+    case 1: mclk = mpll; break;
+    case 2: mclk = mpll / 2.0; break;
+    case 3: mclk = mpll / 4.0; break;
+    case 4: mclk = mpll / 8.0; break;
+    case 7: mclk = spll; break;
+    default:
+           mclk = 200.00;
+           xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Unsupported MCLKA source"
+                      " setting %d, can't probe MCLK value !\n", tmp);
+    }
+
+    /*
+     * Calculate SCLK
+     */
+    tmp = INPLL( RADEON_SCLK_CNTL) & 0x7;
+    switch(tmp) {
+    case 1: sclk = spll; break;
+    case 2: sclk = spll / 2.0; break;
+    case 3: sclk = spll / 4.0; break;
+    case 4: sclk = spll / 8.0; break;
+    case 7: sclk = mpll;
+    default:
+           sclk = 200.00;
+           xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Unsupported SCLK source"
+                      " setting %d, can't probe SCLK value !\n", tmp);
+    }
+
+    /* we're done, hopefully these are sane values */
+    pll->reference_div = ref_div;
+    pll->xclk = xclk;
+    pll->reference_freq = xtal;
+
+    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Probed PLL values: xtal: %f Mhz, "
+              "sclk: %f Mhz, mclk: %f Mhz\n", xtal/100.0, sclk, mclk);
+
+    return 1;
+}
+
+
