@@ -59,13 +59,6 @@ if(vesa_read_write){
 };
 }
 
-
-static int __svgalib_vesa_inlinearmode(void)
-{
-return vesa_is_linear;
-}
-
-
 /* Fill in chipset specific mode information */
 
 static void vesa_getmodeinfo(int mode, vga_modeinfo *modeinfo)
